import pyangle.internals as internals




class tTriangulationParameters(internals.tTriangulationParameters):
    def _constituents(self):
        return [ ("Points", self.Points),
                 ("PointAttributes", self.PointAttributes),
                 ("PointMarkers", self.PointMarkers),
                 
                 ("Triangles", self.Triangles),
                 ("TriangleAttributes", self.TriangleAttributes),
                 ("TriangleAreas", self.TriangleAreas),
                 ("Neighbors", self.Neighbors),
                 
                 ("Segments", self.Segments),
                 ("SegmentMarkers", self.SegmentMarkers),

                 ("Holes", self.Holes),

                 ("Regions", self.Regions),
                 
                 ("Edges", self.Edges),
                 ("EdgeMarkers", self.EdgeMarkers),

                 ("Normals", self.Normals)]

    def __getstate__(self):
        def dumpArray(array):
            try:
                return [
                    [array.getSub(i,j) for j in range(array.unit())]
                    for i in range(array.size())]
            except RuntimeError:
                # not allocated
                return None

        return self.NumberOfPointAttributes, \
               self.NumberOfTriangleAttributes, \
               [(name, dumpArray(arr)) for name, arr in self._constituents()]

    def __setstate__(self, (p_attr_count, t_attr_count, state)):
        self.NumberOfPointAttributes = p_attr_count
        self.NumberOfTriangleAttributes = t_attr_count
        for name, array in state:
            matching_constituents = filter(
                lambda (constit_name, constit_array): name == constit_name,
                self._constituents())
            if len(matching_constituents) != 1:
                raise RuntimeError, "Unknown constituent during unpickling"

            dest_name, dest_array = matching_constituents[0]

            if array is None:
                dest_array.deallocate()
            else:
                if dest_array.size() != len(array):
                    dest_array.setSize(len(array))

                for i,tup in zip(range(len(array)), array):
                    for j,v in zip(range(len(tup)), tup):
                        dest_array.setSub(i, j, v)





def triangulate(input_p, verbose = False, refinement_func = None):
    opts = "pzqj"
    if verbose:
        opts += "VV"
    else:
        opts += "Q"
    if refinement_func is not None:
        opts += "u"
    output_p = tTriangulationParameters()

    internals.triangulate(opts, input_p, output_p, 
                          tTriangulationParameters(), refinement_func)
    return output_p




def refine(input_p, verbose = False, refinement_func = None):
    opts = "razj"
    if input_p.Segments.size() != 0:
        opts += "p"
    if verbose:
        opts += "VV"
    else:
        opts += "Q"
    if refinement_func is not None:
        opts += "u"
    output_p = tTriangulationParameters()
    internals.triangulate(opts, input_p, output_p, 
                          tTriangulationParameters(), refinement_func)
    return output_p




def setPoints(input_p, points, point_markers = None):
    pointcount = len(points)
    
    input_p.Points.setSize(pointcount)

    for i, pt in zip(range(pointcount),points):
        input_p.Points.setSub(i, 0, pt[0])
        input_p.Points.setSub(i, 1, pt[1])
  
    if point_markers is not None:
        assert len(point_markers) == pointcount
        for i, mark in zip(range(pointcount), point_markers):
            input_p.PointMarkers.set(i, mark)




def setSegments(input_p, segments, segment_markers = None):
    input_p.Segments.setSize(len(segments))
    
    for i, seg in zip(range(len(segments)), segments):
        input_p.Segments.setSub(i, 0, seg[0])
        input_p.Segments.setSub(i, 1, seg[1])

    if segment_markers is not None:
        assert len(segment_markers) == len(segments)
        for i, mark in zip(range(len(segment_markers)), segment_markers):
            input_p.SegmentMarkers.set(i, mark)




def setHoles(input_p, hole_starts):
    holecount = len(hole_starts)
    input_p.Holes.setSize(len(hole_starts))
    for i in range(0, holecount):
        input_p.Holes.setSub(i, 0, hole_starts[i][0])
        input_p.Holes.setSub(i, 1, hole_starts[i][1])




def dumpParameters(par):
    def dumpArray(name, array):
        subs = array.unit()
        print "array %s: %d elements, %d values per element" % (name, array.size(), subs)
        if array.size() == 0 or array.unit() == 0:
            return
        try:
            array.get(0)
        except RuntimeError:
            print "  not allocated"
            return

        for i in range(array.size()):
            print "  %d:" % i,
            for j in range(subs):
                print array.getSub(i,j),
            print

        for name, arr in par._constituents():
            dumpArray(name, arr)




def writeGnuplotMesh(gnuplot_file, out_p):
    gp_file = file(gnuplot_file, "w")
    
    pts = out_p.Points
    tris = out_p.Triangles

    for tri in range(out_p.Triangles.size()):
        for pt in range(0, 3):
            gp_file.write("%f %f 0\n" % (
                pts.getSub(tris.getSub(tri, pt), 0),
                pts.getSub(tris.getSub(tri, pt), 1)))
        gp_file.write("%f %f 0\n\n" % (
            pts.getSub(tris.getSub(tri, 0), 0),
            pts.getSub(tris.getSub(tri, 0), 1)))

